"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.convertAttributes = void 0;
const common_1 = require("./common");
const es_1 = require("./es");
const mustache_1 = require("./mustache");
const text_1 = require("./text");
const errors_1 = require("../../errors");
/** Convert for Attributes */
function* convertAttributes(attributes, parent, ctx) {
    for (const attr of attributes) {
        if (attr.type === "Attribute") {
            yield convertAttribute(attr, parent, ctx);
            continue;
        }
        if (attr.type === "Spread") {
            yield convertSpreadAttribute(attr, parent, ctx);
            continue;
        }
        if (attr.type === "Binding") {
            yield convertBindingDirective(attr, parent, ctx);
            continue;
        }
        if (attr.type === "EventHandler") {
            yield convertEventHandlerDirective(attr, parent, ctx);
            continue;
        }
        if (attr.type === "Class") {
            yield convertClassDirective(attr, parent, ctx);
            continue;
        }
        if (attr.type === "Transition") {
            yield convertTransitionDirective(attr, parent, ctx);
            continue;
        }
        if (attr.type === "Animation") {
            yield convertAnimationDirective(attr, parent, ctx);
            continue;
        }
        if (attr.type === "Action") {
            yield convertActionDirective(attr, parent, ctx);
            continue;
        }
        if (attr.type === "Let") {
            yield convertLetDirective(attr, parent, ctx);
            continue;
        }
        if (attr.type === "Ref") {
            throw new errors_1.ParseError("Ref are not supported.", attr.start, ctx);
        }
        throw new errors_1.ParseError(`Unknown directive or attribute (${attr.type}) are not supported.`, attr.start, ctx);
    }
}
exports.convertAttributes = convertAttributes;
/** Convert for Attribute */
function convertAttribute(node, parent, ctx) {
    const attribute = Object.assign({ type: "SvelteAttribute", boolean: false, key: null, value: [], parent }, ctx.getConvertLocation(node));
    const keyStart = ctx.code.indexOf(node.name, node.start);
    const keyRange = { start: keyStart, end: keyStart + node.name.length };
    attribute.key = Object.assign({ type: "SvelteName", name: node.name, parent: attribute }, ctx.getConvertLocation(keyRange));
    ctx.addToken("HTMLIdentifier", keyRange);
    if (node.value === true) {
        // Boolean attribute
        attribute.boolean = true;
        return attribute;
    }
    for (const v of node.value) {
        if (v.type === "AttributeShorthand") {
            const key = Object.assign(Object.assign({}, attribute.key), { type: "Identifier" });
            const sAttr = {
                type: "SvelteShorthandAttribute",
                key,
                value: key,
                parent,
                loc: attribute.loc,
                range: attribute.range,
            };
            key.parent = sAttr;
            es_1.analyzeExpressionScope(sAttr.key, ctx);
            return sAttr;
        }
        if (v.type === "Text") {
            attribute.value.push(text_1.convertTextToLiteral(v, attribute, ctx));
            continue;
        }
        if (v.type === "MustacheTag") {
            const mustache = mustache_1.convertMustacheTag(v, attribute, ctx);
            attribute.value.push(mustache);
            continue;
        }
        const u = v;
        throw new errors_1.ParseError(`Unknown attribute value (${u.type}) are not supported.`, u.start, ctx);
    }
    return attribute;
}
/** Convert for Spread */
function convertSpreadAttribute(node, parent, ctx) {
    const attribute = Object.assign({ type: "SvelteSpreadAttribute", argument: null, parent }, ctx.getConvertLocation(node));
    const spreadStart = ctx.code.indexOf("...", node.start);
    ctx.addToken("Punctuator", {
        start: spreadStart,
        end: spreadStart + 3,
    });
    const es = es_1.convertESNode(node.expression, attribute, ctx);
    es_1.analyzeExpressionScope(es, ctx);
    attribute.argument = es;
    return attribute;
}
/** Convert for Binding Directive */
function convertBindingDirective(node, parent, ctx) {
    const directive = Object.assign({ type: "SvelteDirective", kind: "Binding", name: null, modifiers: node.modifiers, expression: null, parent }, ctx.getConvertLocation(node));
    processDirective(node, directive, ctx, (expression) => {
        const es = es_1.convertESNode(expression, directive, ctx);
        es_1.analyzeExpressionScope(es, ctx);
        const reference = ctx.templateScopeManager.currentScope.references.find((ref) => ref.identifier === es);
        if (reference) {
            // The bind directive does read and write.
            reference.isWrite = () => true;
            reference.isWriteOnly = () => false;
            reference.isReadWrite = () => true;
            reference.isReadOnly = () => false;
            reference.isRead = () => true;
        }
        return es;
    });
    return directive;
}
/** Convert for EventHandler Directive */
function convertEventHandlerDirective(node, parent, ctx) {
    const directive = Object.assign({ type: "SvelteDirective", kind: "EventHandler", name: null, modifiers: node.modifiers, expression: null, parent }, ctx.getConvertLocation(node));
    processDirective(node, directive, ctx, buildProcessExpressionForExpression(directive, ctx));
    return directive;
}
/** Convert for Class Directive */
function convertClassDirective(node, parent, ctx) {
    const directive = Object.assign({ type: "SvelteDirective", kind: "Class", name: null, modifiers: node.modifiers, expression: null, parent }, ctx.getConvertLocation(node));
    processDirective(node, directive, ctx, buildProcessExpressionForExpression(directive, ctx));
    return directive;
}
/** Convert for Transition Directive */
function convertTransitionDirective(node, parent, ctx) {
    const directive = Object.assign({ type: "SvelteDirective", kind: "Transition", intro: node.intro, outro: node.outro, name: null, modifiers: node.modifiers, expression: null, parent }, ctx.getConvertLocation(node));
    processDirective(node, directive, ctx, buildProcessExpressionForExpression(directive, ctx));
    if (directive.expression !== directive.name) {
        // Transition name is reference
        es_1.analyzeExpressionScope(directive.name, ctx);
    }
    return directive;
}
/** Convert for Animation Directive */
function convertAnimationDirective(node, parent, ctx) {
    const directive = Object.assign({ type: "SvelteDirective", kind: "Animation", name: null, modifiers: node.modifiers, expression: null, parent }, ctx.getConvertLocation(node));
    processDirective(node, directive, ctx, buildProcessExpressionForExpression(directive, ctx));
    if (directive.expression !== directive.name) {
        // Animation name is reference
        es_1.analyzeExpressionScope(directive.name, ctx);
    }
    return directive;
}
/** Convert for Action Directive */
function convertActionDirective(node, parent, ctx) {
    const directive = Object.assign({ type: "SvelteDirective", kind: "Action", name: null, modifiers: node.modifiers, expression: null, parent }, ctx.getConvertLocation(node));
    processDirective(node, directive, ctx, buildProcessExpressionForExpression(directive, ctx));
    if (directive.expression !== directive.name) {
        // Action name is reference
        es_1.analyzeExpressionScope(directive.name, ctx);
    }
    return directive;
}
/** Convert for Let Directive */
function convertLetDirective(node, parent, ctx) {
    const directive = Object.assign({ type: "SvelteDirective", kind: "Let", name: null, modifiers: node.modifiers, expression: null, parent }, ctx.getConvertLocation(node));
    processDirective(node, directive, ctx, (pattern) => {
        const es = es_1.convertESNode(pattern, directive, ctx);
        es_1.analyzePatternScope(es, ctx);
        return es;
    });
    if (!directive.expression) {
        // shorthand
        es_1.analyzePatternScope(directive.name, ctx);
        directive.expression = directive.name;
    }
    return directive;
}
/** Common process for directive */
function processDirective(node, directive, ctx, processExpression) {
    var _a;
    const colonIndex = ctx.code.indexOf(":", directive.range[0]);
    ctx.addToken("HTMLIdentifier", {
        start: directive.range[0],
        end: colonIndex,
    });
    const nameIndex = ctx.code.indexOf(node.name, colonIndex + 1);
    const nameRange = {
        start: nameIndex,
        end: nameIndex + node.name.length,
    };
    // modifiers
    if (ctx.code[nameRange.end] === "|") {
        let nextStart = nameRange.end + 1;
        let nextEnd = common_1.indexOf(ctx.code, (c) => c === "=" || c === ">" || c === "|" || !c.trim(), nextStart);
        ctx.addToken("HTMLIdentifier", { start: nextStart, end: nextEnd });
        while (ctx.code[nextEnd] === "|") {
            nextStart = nextEnd + 1;
            nextEnd = common_1.indexOf(ctx.code, (c) => c === "=" || c === ">" || c === "|" || !c.trim(), nextStart);
            ctx.addToken("HTMLIdentifier", { start: nextStart, end: nextEnd });
        }
    }
    if (node.expression) {
        directive.expression = processExpression(node.expression);
    }
    // put name
    if (((_a = directive.expression) === null || _a === void 0 ? void 0 : _a.type) === "Identifier" &&
        directive.expression.name === node.name &&
        es_1.getWithLoc(directive.expression).start === nameRange.start &&
        es_1.getWithLoc(directive.expression).end === nameRange.end) {
        // shorthand
        directive.name = directive.expression;
    }
    else {
        directive.name = Object.assign({ type: "Identifier", name: node.name, 
            // @ts-expect-error -- ignore
            parent: directive }, ctx.getConvertLocation(nameRange));
        ctx.addToken("HTMLIdentifier", nameRange);
    }
}
/** Build processExpression for Expression */
function buildProcessExpressionForExpression(directive, ctx) {
    return (expression) => {
        const es = es_1.convertESNode(expression, directive, ctx);
        es_1.analyzeExpressionScope(es, ctx);
        return es;
    };
}
