"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getWithLoc = exports.analyzePatternScope = exports.analyzeExpressionScope = exports.convertESNode = void 0;
const traverse_1 = require("../../traverse");
const analyze_scope_1 = require("../analyze-scope");
/** Add operator token */
function addToken(index, text, ctx) {
    const range = {
        start: index,
        end: index + text.length,
    };
    if (text === "of" ||
        text === "async" ||
        text === "await" ||
        text === "as" ||
        text === "from" ||
        text === "get" ||
        text === "set") {
        ctx.addToken("Identifier", range);
    }
    else if (/^[a-z]+$/i.test(text)) {
        ctx.addToken("Keyword", range);
    }
    else {
        ctx.addToken("Punctuator", range);
    }
}
const EXTRACT_TOKENS0 = {
    ArrayExpression() {
        /* noop */
    },
    ArrayPattern() {
        /* noop */
    },
    ArrowFunctionExpression(node, ctx) {
        var _a, _b;
        const index = ctx.code.indexOf("=>", (_b = (_a = getWithLoc(node.params[node.params.length - 1])) === null || _a === void 0 ? void 0 : _a.end) !== null && _b !== void 0 ? _b : getWithLoc(node).start);
        addToken(index, "=>", ctx);
    },
    AssignmentExpression(node, ctx) {
        const operator = node.type === "AssignmentPattern" ? "=" : node.operator;
        const index = ctx.code.indexOf(operator, getWithLoc(node.left).end);
        addToken(index, operator, ctx);
    },
    AssignmentPattern(node, ctx) {
        EXTRACT_TOKENS0.AssignmentExpression(node, ctx);
    },
    AwaitExpression(node, ctx) {
        const index = ctx.code.indexOf("await", getWithLoc(node).start);
        addToken(index, "await", ctx);
    },
    BinaryExpression(node, ctx) {
        EXTRACT_TOKENS0.AssignmentExpression(node, ctx);
    },
    BlockStatement() {
        /* noop */
    },
    BreakStatement(node, ctx) {
        const index = ctx.code.indexOf("break", getWithLoc(node).start);
        addToken(index, "break", ctx);
    },
    CallExpression(node, ctx) {
        if (node.optional) {
            const index = ctx.code.indexOf("?.", getWithLoc(node.callee).end);
            addToken(index, "?.", ctx);
        }
    },
    CatchClause(node, ctx) {
        const index = ctx.code.indexOf("catch", getWithLoc(node).start);
        addToken(index, "catch", ctx);
    },
    ChainExpression() {
        /* noop */
    },
    ClassBody() {
        /* noop */
    },
    ClassDeclaration(node, ctx) {
        const classIndex = ctx.code.indexOf("class", getWithLoc(node).start);
        addToken(classIndex, "class", ctx);
        if (node.superClass) {
            const extendsIndex = ctx.code.lastIndexOf("extends", getWithLoc(node.superClass).start);
            addToken(extendsIndex, "extends", ctx);
        }
    },
    ClassExpression(node, ctx) {
        EXTRACT_TOKENS0.ClassDeclaration(node, ctx);
    },
    ConditionalExpression() {
        /* noop */
    },
    ContinueStatement(node, ctx) {
        const index = ctx.code.indexOf("continue", getWithLoc(node).start);
        addToken(index, "continue", ctx);
    },
    DebuggerStatement(node, ctx) {
        const index = ctx.code.indexOf("debugger", getWithLoc(node).start);
        addToken(index, "debugger", ctx);
    },
    DoWhileStatement(node, ctx) {
        const doIndex = ctx.code.indexOf("do", getWithLoc(node).start);
        addToken(doIndex, "do", ctx);
        const whileIndex = ctx.code.lastIndexOf("while", getWithLoc(node.test).start);
        addToken(whileIndex, "while", ctx);
    },
    EmptyStatement() {
        /* noop */
    },
    ExportAllDeclaration(node, ctx) {
        const exportIndex = ctx.code.indexOf("export", getWithLoc(node).start);
        addToken(exportIndex, "export", ctx);
        if (node.exported) {
            const asIndex = ctx.code.lastIndexOf("as", getWithLoc(node.exported).start);
            addToken(asIndex, "as", ctx);
        }
        const fromIndex = ctx.code.lastIndexOf("from", getWithLoc(node.source).start);
        addToken(fromIndex, "from", ctx);
    },
    ExportDefaultDeclaration(node, ctx) {
        const exportIndex = ctx.code.indexOf("export", getWithLoc(node).start);
        addToken(exportIndex, "export", ctx);
        const defaultIndex = ctx.code.indexOf("default", exportIndex + 6);
        addToken(defaultIndex, "default", ctx);
    },
    ExportNamedDeclaration(node, ctx) {
        const exportIndex = ctx.code.indexOf("export", getWithLoc(node).start);
        addToken(exportIndex, "export", ctx);
    },
    ExportSpecifier(node, ctx) {
        if (node.local !== node.exported) {
            const asIndex = ctx.code.indexOf("as", getWithLoc(node.local).end);
            addToken(asIndex, "as", ctx);
        }
    },
    ExpressionStatement() {
        /* noop */
    },
    ForInStatement(node, ctx) {
        const forIndex = ctx.code.indexOf("for", getWithLoc(node).start);
        addToken(forIndex, "for", ctx);
        if (node.type === "ForOfStatement" && node.await) {
            const awaitIndex = ctx.code.indexOf("await", forIndex + 3);
            addToken(awaitIndex, "await", ctx);
        }
        if (node.type !== "ForStatement") {
            const keyword = node.type === "ForInStatement" ? "in" : "of";
            const keywordIndex = ctx.code.indexOf(keyword, getWithLoc(node.left).end);
            addToken(keywordIndex, keyword, ctx);
        }
    },
    ForOfStatement(node, ctx) {
        EXTRACT_TOKENS0.ForInStatement(node, ctx);
    },
    ForStatement(node, ctx) {
        EXTRACT_TOKENS0.ForInStatement(node, ctx);
    },
    FunctionDeclaration(node, ctx) {
        let start = getWithLoc(node).start;
        if (node.async) {
            const index = ctx.code.indexOf("async", start);
            addToken(index, "async", ctx);
            start = index + 5;
        }
        const index = ctx.code.indexOf("function", start);
        addToken(index, "function", ctx);
    },
    FunctionExpression(node, ctx, parent) {
        if (parent &&
            (parent.type === "MethodDefinition" ||
                (parent.type === "Property" &&
                    parent.value === node &&
                    (parent.method ||
                        parent.kind === "get" ||
                        parent.kind === "set")))) {
            return;
        }
        EXTRACT_TOKENS0.FunctionDeclaration(node, ctx);
    },
    Identifier(node, ctx, parent) {
        if ((parent === null || parent === void 0 ? void 0 : parent.type) === "MetaProperty" && parent.meta === node) {
            ctx.addToken("Keyword", getWithLoc(node));
        }
        else {
            ctx.addToken("Identifier", getWithLoc(node));
        }
    },
    IfStatement(node, ctx) {
        const index = ctx.code.indexOf("if", getWithLoc(node).start);
        addToken(index, "if", ctx);
        if (node.alternate) {
            const elseIndex = ctx.code.lastIndexOf("else", getWithLoc(node.alternate).start);
            addToken(elseIndex, "else", ctx);
        }
    },
    ImportDeclaration(node, ctx) {
        const importIndex = ctx.code.indexOf("import", getWithLoc(node).start);
        addToken(importIndex, "import", ctx);
        if (node.specifiers.length) {
            const fromIndex = ctx.code.lastIndexOf("from", getWithLoc(node.source).start);
            addToken(fromIndex, "from", ctx);
        }
    },
    ImportDefaultSpecifier() {
        /* noop */
    },
    ImportExpression(node, ctx) {
        const importIndex = ctx.code.indexOf("import", getWithLoc(node).start);
        addToken(importIndex, "import", ctx);
    },
    ImportNamespaceSpecifier(node, ctx) {
        const asIndex = ctx.code.lastIndexOf("as", getWithLoc(node.local).end);
        addToken(asIndex, "as", ctx);
    },
    ImportSpecifier(node, ctx) {
        if (node.local !== node.imported) {
            const asIndex = ctx.code.indexOf("as", getWithLoc(node.imported).end);
            addToken(asIndex, "as", ctx);
        }
    },
    LabeledStatement() {
        /* noop */
    },
    Literal(node, ctx) {
        if ("regex" in node) {
            const token = ctx.addToken("RegularExpression", getWithLoc(node));
            token.regex = node.regex;
        }
        else if ("bigint" in node) {
            ctx.addToken("Numeric", getWithLoc(node));
        }
        else if (node.value === null) {
            ctx.addToken("Null", getWithLoc(node));
        }
        else if (typeof node.value === "number") {
            ctx.addToken("Numeric", getWithLoc(node));
        }
        else if (typeof node.value === "string") {
            ctx.addToken("String", getWithLoc(node));
        }
        else if (typeof node.value === "boolean") {
            ctx.addToken("Boolean", getWithLoc(node));
        }
    },
    LogicalExpression(node, ctx) {
        EXTRACT_TOKENS0.AssignmentExpression(node, ctx);
    },
    MemberExpression(node, ctx) {
        if (node.optional) {
            const index = ctx.code.indexOf("?.", getWithLoc(node.object).end);
            addToken(index, "?.", ctx);
        }
    },
    MetaProperty() {
        /* noop */
    },
    MethodDefinition(node, ctx) {
        let start = getWithLoc(node).start;
        if (node.static) {
            const index = ctx.code.indexOf("static", start);
            addToken(index, "static", ctx);
            start = index + 6;
        }
        if (node.value.async) {
            const index = ctx.code.indexOf("async", start);
            addToken(index, "async", ctx);
            start = index + 5;
        }
        if (node.kind === "get" || node.kind === "set") {
            const index = ctx.code.indexOf(node.kind, start);
            addToken(index, node.kind, ctx);
        }
    },
    NewExpression(node, ctx) {
        const index = ctx.code.indexOf("new", getWithLoc(node).start);
        addToken(index, "new", ctx);
    },
    ObjectExpression() {
        /* noop */
    },
    ObjectPattern() {
        /* noop */
    },
    Program() {
        /* noop */
    },
    Property(node, ctx) {
        const start = getWithLoc(node).start;
        if (node.kind === "get" || node.kind === "set") {
            const index = ctx.code.indexOf(node.kind, start);
            addToken(index, node.kind, ctx);
        }
    },
    RestElement(node, ctx) {
        const index = ctx.code.indexOf("...", getWithLoc(node).start);
        addToken(index, "...", ctx);
    },
    ReturnStatement(node, ctx) {
        const index = ctx.code.indexOf("return", getWithLoc(node).start);
        addToken(index, "return", ctx);
    },
    SequenceExpression() {
        /* noop */
    },
    SpreadElement(node, ctx) {
        EXTRACT_TOKENS0.RestElement(node, ctx);
    },
    Super(node, ctx) {
        const index = ctx.code.indexOf("super", getWithLoc(node).start);
        addToken(index, "super", ctx);
    },
    SwitchCase(node, ctx) {
        if (node.test) {
            const index = ctx.code.indexOf("case", getWithLoc(node).start);
            addToken(index, "case", ctx);
        }
        else {
            const index = ctx.code.indexOf("default", getWithLoc(node).start);
            addToken(index, "default", ctx);
        }
    },
    SwitchStatement(node, ctx) {
        const index = ctx.code.indexOf("switch", getWithLoc(node).start);
        addToken(index, "switch", ctx);
    },
    TaggedTemplateExpression() {
        /* noop */
    },
    TemplateElement(node, ctx, parent) {
        const literal = parent;
        const start = literal.quasis[0] === node
            ? getWithLoc(literal).start
            : getWithLoc(node).start - 1;
        const end = literal.quasis[literal.quasis.length - 1] === node
            ? getWithLoc(literal).end
            : getWithLoc(node).end + 2;
        ctx.addToken("Template", { start, end });
    },
    TemplateLiteral() {
        /* noop */
    },
    ThisExpression(node, ctx) {
        const index = ctx.code.indexOf("this", getWithLoc(node).start);
        addToken(index, "this", ctx);
    },
    ThrowStatement(node, ctx) {
        const index = ctx.code.indexOf("throw", getWithLoc(node).start);
        addToken(index, "throw", ctx);
    },
    TryStatement(node, ctx) {
        const index = ctx.code.indexOf("try", getWithLoc(node).start);
        addToken(index, "try", ctx);
    },
    UnaryExpression(node, ctx) {
        const index = ctx.code.lastIndexOf(node.operator, getWithLoc(node.argument).start - 1);
        addToken(index, node.operator, ctx);
    },
    UpdateExpression(node, ctx) {
        const index = node.prefix
            ? ctx.code.lastIndexOf(node.operator, getWithLoc(node.argument).start - 1)
            : ctx.code.indexOf(node.operator, getWithLoc(node.argument).end);
        addToken(index, node.operator, ctx);
    },
    VariableDeclaration(node, ctx) {
        const index = ctx.code.indexOf(node.kind, getWithLoc(node).start);
        addToken(index, node.kind, ctx);
    },
    VariableDeclarator() {
        /* noop */
    },
    WhileStatement(node, ctx) {
        const index = ctx.code.indexOf("while", getWithLoc(node).start);
        addToken(index, "while", ctx);
    },
    WithStatement(node, ctx) {
        const index = ctx.code.indexOf("with", getWithLoc(node).start);
        addToken(index, "with", ctx);
    },
    YieldExpression(node, ctx) {
        const index = ctx.code.indexOf("yield", getWithLoc(node).start);
        addToken(index, "yield", ctx);
    },
};
const EXTRACT_TOKENS = EXTRACT_TOKENS0;
const GET_LOC = {
    ArrayExpression: getWithLoc,
    ArrayPattern: getWithLoc,
    ArrowFunctionExpression: getWithLoc,
    AssignmentExpression: getWithLoc,
    AssignmentPattern: getWithLoc,
    AwaitExpression: getWithLoc,
    BinaryExpression: getWithLoc,
    BlockStatement: getWithLoc,
    BreakStatement: getWithLoc,
    CallExpression: getWithLoc,
    CatchClause: getWithLoc,
    ChainExpression: getWithLoc,
    ClassBody: getWithLoc,
    ClassDeclaration: getWithLoc,
    ClassExpression: getWithLoc,
    ConditionalExpression: getWithLoc,
    ContinueStatement: getWithLoc,
    DebuggerStatement: getWithLoc,
    DoWhileStatement: getWithLoc,
    EmptyStatement: getWithLoc,
    ExportAllDeclaration: getWithLoc,
    ExportDefaultDeclaration: getWithLoc,
    ExportNamedDeclaration: getWithLoc,
    ExportSpecifier: getWithLoc,
    ExpressionStatement: getWithLoc,
    ForInStatement: getWithLoc,
    ForOfStatement: getWithLoc,
    ForStatement: getWithLoc,
    FunctionDeclaration: getWithLoc,
    FunctionExpression: getWithLoc,
    Identifier: getWithLoc,
    IfStatement: getWithLoc,
    ImportDeclaration: getWithLoc,
    ImportDefaultSpecifier: getWithLoc,
    ImportExpression: getWithLoc,
    ImportNamespaceSpecifier: getWithLoc,
    ImportSpecifier: getWithLoc,
    LabeledStatement: getWithLoc,
    Literal: getWithLoc,
    LogicalExpression: getWithLoc,
    MemberExpression: getWithLoc,
    MetaProperty: getWithLoc,
    MethodDefinition: getWithLoc,
    NewExpression: getWithLoc,
    ObjectExpression: getWithLoc,
    ObjectPattern: getWithLoc,
    Program: getWithLoc,
    Property: getWithLoc,
    RestElement: getWithLoc,
    ReturnStatement: getWithLoc,
    SequenceExpression: getWithLoc,
    SpreadElement: getWithLoc,
    Super: getWithLoc,
    SwitchCase: getWithLoc,
    SwitchStatement: getWithLoc,
    TaggedTemplateExpression: getWithLoc,
    TemplateElement(node, _ctx, parent) {
        const literal = parent;
        const start = literal.quasis[0] === node
            ? getWithLoc(literal).start
            : getWithLoc(node).start - 1;
        const end = literal.quasis[literal.quasis.length - 1] === node
            ? getWithLoc(literal).end
            : getWithLoc(node).end + 2;
        return { start, end };
    },
    TemplateLiteral: getWithLoc,
    ThisExpression: getWithLoc,
    ThrowStatement: getWithLoc,
    TryStatement: getWithLoc,
    UnaryExpression: getWithLoc,
    UpdateExpression: getWithLoc,
    VariableDeclaration: getWithLoc,
    VariableDeclarator: getWithLoc,
    WhileStatement: getWithLoc,
    WithStatement: getWithLoc,
    YieldExpression: getWithLoc,
};
/** Convert for ES */
function convertESNode(node, parent, ctx) {
    return convertESNode0(node, parent, ctx);
}
exports.convertESNode = convertESNode;
/** Convert for ES */
function convertESNode0(node, parent, ctx, extracted = new Set()) {
    var _a, _b;
    if (!node) {
        return null;
    }
    if (!extracted.has(node)) {
        const key = `${node.type}@[${getWithLoc(node).start},${getWithLoc(node).end}]`;
        if (!extracted.has(key)) {
            (_a = EXTRACT_TOKENS[node.type]) === null || _a === void 0 ? void 0 : _a.call(EXTRACT_TOKENS, node, ctx, parent);
            for (const comment of [
                ...(node.leadingComments || []),
                ...(node.innerComments || []),
                ...(node.trailingComments || []),
            ]) {
                ctx.addComment(Object.assign({ type: comment.type, value: comment.value }, ctx.getConvertLocation(getWithLoc(comment))));
            }
            extracted.add(key);
        }
        extracted.add(node);
    }
    const result = Object.assign(Object.assign(Object.assign({}, node), { parent }), ctx.getConvertLocation(((_b = GET_LOC[node.type]) !== null && _b !== void 0 ? _b : getWithLoc)(node, ctx, parent)));
    for (const key of traverse_1.getKeys(node)) {
        const child = node[key];
        result[key] = Array.isArray(child)
            ? child.map((c) => convertESNode0(c, result, ctx, extracted))
            : convertESNode0(child, result, ctx, extracted);
    }
    return result;
}
/** Analyze expression scope */
function analyzeExpressionScope(node, ctx) {
    const scopeManager = analyze_scope_1.analyzeScope(node, ctx.parserOptions);
    const moduleScope = scopeManager.globalScope.childScopes[0];
    // Merge
    mergeScope(moduleScope, ctx);
}
exports.analyzeExpressionScope = analyzeExpressionScope;
/** Analyze pattern scope */
function analyzePatternScope(node, ctx) {
    const fnNode = {
        type: "FunctionExpression",
        body: { type: "BlockStatement", body: [] },
        params: [node],
    };
    const scopeManager = analyze_scope_1.analyzeScope(fnNode, ctx.parserOptions);
    const moduleScope = scopeManager.globalScope.childScopes[0];
    const forScope = moduleScope.childScopes[0];
    forScope.variables = forScope.variables.filter((variable) => {
        if (variable.name === "arguments") {
            return false;
        }
        for (const def of variable.defs) {
            if (def.node === fnNode) {
                def.node = node.parent;
            }
        }
        return true;
    });
    // Merge
    mergeScope(forScope, ctx);
}
exports.analyzePatternScope = analyzePatternScope;
/** Merge scope */
function mergeScope(targetScope, ctx) {
    const templateScopeManager = ctx.templateScopeManager;
    for (const scope of targetScope.childScopes) {
        templateScopeManager.registerScope(scope);
        templateScopeManager.currentScope.childScopes.push(scope);
        scope.upper = templateScopeManager.currentScope;
    }
    for (const variable of targetScope.variables) {
        ;
        variable.scope = templateScopeManager.currentScope;
        templateScopeManager.currentScope.variables.push(variable);
        templateScopeManager.currentScope.set.set(variable.name, variable);
    }
    for (const reference of targetScope.references) {
        reference.from = templateScopeManager.currentScope;
        templateScopeManager.currentScope.references.push(reference);
    }
    // Merge through
    for (const reference of targetScope.through) {
        let scope = templateScopeManager.currentScope;
        let variable = scope.set.get(reference.identifier.name);
        while (!variable) {
            scope.through.push(reference);
            if (scope.upper) {
                scope = scope.upper;
                variable = scope.set.get(reference.identifier.name);
                continue;
            }
            break;
        }
        if (variable) {
            reference.resolved = variable;
            variable.references.push(reference);
        }
    }
}
/** Get node with location */
function getWithLoc(node) {
    return node;
}
exports.getWithLoc = getWithLoc;
